#ifndef _TYPES_COMPAT_H_
#define	_TYPES_COMPAT_H_

#include <stdint.h>
#include <limits.h>

typedef uint8_t               u_int8_t;
typedef uint16_t              u_int16_t;
typedef uint32_t              u_int32_t;
typedef uint64_t              u_int64_t;


#endif
